from typing import Tuple

import torch
from sklearn.metrics import matthews_corrcoef, r2_score
from torch import Tensor, nn
from torch_geometric.loader import DataLoader

from MolGNN import models
from MolGNN.utils import logger


class GraphLevelGNN(nn.Module):

    def __init__(self, device: str = 'cpu', **model_kwargs):
        """
        Args:
            device (str)
            **model_kwargs:

        This is a base class that wraps around a GraphNNModel
        that uses GNNModel layers

        GraphGNNModel level args:
            dp_rate_linear (float): Dropout rate before the linear layer (usually much higher than inside the GNN)
            device (str)

        GNNModel level args:
            c_in (int): Dimension of input features
            c_hidden (int): Dimension of hidden features
            c_out (int): Dimension of the output features. Usually number of classes in classification
            num_layers (int): Number of "hidden" graph layers
            layer_name (str): String of the graph layer to use (see gnn_layer_by_name dict)
            dp_rate (float): Dropout rate to apply throughout the network

        Attributes:
            model (nn.Model): Underlying GraphGNN model
            optimizer: Torch optimizer
            criterion: Torch loss module
            metric: Callable that takes the output of the forward pass and the batch labels and
                returns a metric value
            logger: Logger
        """

        super().__init__()

        self.device = device
        self.model = models.GraphGNNModel(device=device, **model_kwargs).to(device)

        self.optimizer = None
        self.criterion = None
        self.metric = None

        self.logger = logger

    def forward(self, batch) -> Tensor:
        """
        Forward pass

        Args:
            batch (DataBatch): batch of PyTorch geometric dataloader

        Returns:
            outputs (Tensor)
        """
        x, edge_index, batch_idx = batch.x, batch.edge_index, batch.batch
        outputs = self.model(x, edge_index, batch_idx)
        outputs = outputs.squeeze(dim=-1)
        return outputs

    def step(self, batch) -> Tensor:
        """
        Perform one training step: forward pass, loss evaluation, optimizer step

        Args:
            batch (DataBatch): batch of PyTorch geometric dataloader

        Returns:
            loss (Tensor)
        """
        self.optimizer.zero_grad()

        outputs = self.forward(batch)

        loss = self.criterion(outputs, batch.y)
        loss.backward()
        self.optimizer.step()

        return loss

    def fit(self, train_loader: DataLoader, val_loader: DataLoader, n_epochs: int, log_every_epochs: int) -> Tuple[
        Tensor, float]:
        """
        Train the model

        Args:
            train_loader (DataLoader)
            val_loader (DataLoader)
            n_epochs (int)
            log_every_epochs (int)

        Returns:
            val_loss (Tensor)
            val_metric (float)
        """

        for epoch in range(n_epochs):
            for batch in train_loader:
                _ = self.step(batch)

            if (epoch % log_every_epochs == 0) or (epoch == n_epochs - 1):
                val_loss, val_metric = self.evaluate(val_loader)

                self.logger.info(
                    f"Epoch: {epoch + 1:3d}/{n_epochs:3d} |"
                    f" val loss: {val_loss:8.3f} | val metric: {val_metric:8.3f}"
                )

        val_loss, val_metric = self.evaluate(val_loader)
        return val_loss, val_metric

    def evaluate(self, loader):
        """
        Evaluate the model on a validation/test set

        Args:
            loader (DataLoader)

        Returns:
            val_loss (Tensor)
            val_metric (float)
        """

        train_mode = self.model.training

        if train_mode:
            self.model.eval()

        with torch.no_grad():
            val_outputs, val_y = [], []
            for val_batch in loader:
                val_outputs.append(self.forward(val_batch))
                val_y.append(val_batch.y)

        val_outputs = torch.cat(val_outputs).cpu()
        val_y = torch.cat(val_y).cpu()

        val_loss = self.criterion(val_outputs, val_y)
        val_metric = self.metric(val_y, val_outputs)

        if train_mode:
            self.model.train()

        return val_loss, val_metric


class GraphClassifier(GraphLevelGNN):

    def __init__(self, num_classes, lr: float, weight_decay: float, device: str = 'cpu', **model_kwargs):
        """
        GNN Classifier

        Optimizer: Adam
        Criterion: BCEWithLogsLoss or CrossEntropyLoss

        Args:
            num_classes (int)
            lr (float): Adam lr
            weight_decay (float): Adam weight_decay
            device (str)
            **model_kwargs:
        """

        super().__init__(c_out=num_classes, device=device, **model_kwargs)

        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay=weight_decay)
        self.criterion = nn.BCEWithLogitsLoss() if num_classes == 1 else nn.CrossEntropyLoss()
        self.metric = matthews_corrcoef

    def evaluate(self, loader, threshold: float = 0.5):
        """
        Evaluate the model on a validation/test set

        Args:
            loader (DataLoader)
            threshold (float): Classification threshold

        Returns:
            val_loss (Tensor)
            val_metric (float)
        """

        train_mode = self.model.training

        if train_mode:
            self.model.eval()

        with torch.no_grad():
            val_outputs, val_y = [], []
            for val_batch in loader:
                val_outputs.append(self.forward(val_batch))
                val_y.append(val_batch.y)

        val_outputs = torch.cat(val_outputs).cpu()
        val_y = torch.cat(val_y).cpu()

        preds = (val_outputs >= threshold).to(int)

        val_loss = self.criterion(val_outputs, val_y)
        val_metric = self.metric(val_y, preds)

        if train_mode:
            self.model.train()

        return val_loss, val_metric


class GraphRegressor(GraphLevelGNN):

    def __init__(self, lr: float, weight_decay: float, device: str = 'cpu', **model_kwargs):
        """
        GNN Regressor

        Optimizer: Adam
        Criterion: MSELoss

        Args:
            num_classes (int)
            lr (float): Adam lr
            weight_decay (float): Adam weight_decay
            device (str)
            **model_kwargs:
        """

        super().__init__(c_out=1, device=device, **model_kwargs)

        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay=weight_decay)
        self.criterion = nn.MSELoss()
        self.metric = r2_score



import torch_geometric.nn as geom_nn
from torch import nn

from MolGNN import models


class GraphGNNModel(nn.Module):

    def __init__(self, c_in: int, c_hidden: int, c_out: int, dp_rate_linear: float = 0.5, device: str = 'cpu',
                 **kwargs):
        """
        Graph GCN model

        Args:
            c_in (int): Dimension of input features
            c_hidden (int): Dimension of hidden features
            c_out (int): Dimension of output features (usually number of classes)
            dp_rate_linear (float): Dropout rate before the linear layer (usually much higher than inside the GNN)
            device (str)
            **kwargs: Additional arguments for the GNNModel object
        """

        super().__init__()
        self.GNN = models.GNNModel(c_in=c_in,
                                   c_hidden=c_hidden,
                                   c_out=c_hidden,  # Not our prediction output yet!
                                   **kwargs).to(device)
        self.head = nn.Sequential(
            nn.Dropout(dp_rate_linear),
            nn.Linear(c_hidden, c_out)
        )

    def forward(self, x, edge_index, batch_idx):
        """
        Args:
            x: Input features per node
            edge_index: List of vertex index pairs representing the edges in the graph (PyTorch geometric notation)
            batch_idx: Index of batch element for each node

        Returns:
            x: Output
        """
        x = self.GNN(x, edge_index)
        x = geom_nn.global_mean_pool(x, batch_idx)  # Average pooling
        x = self.head(x)
        return x



from numbers import Number
from typing import Sequence, Any, List

import numpy as np
import torch
from rdkit import Chem
from rdkit.Chem.rdmolops import GetAdjacencyMatrix
from torch_geometric.data import Data


def one_hot_encoding(x: Any, values: Sequence[Any]) -> List[int]:
    """
    Sparse one-hot encoding of an input value, given a list of possible
    values. If x is not in values, an extra dimension is added to the vector and
    set to 1

    Args:
        x (Any)
        values (Sequence[Any]): Possible values

    Returns:
        binary_encoding (List[int]): Sparse one-hot vector

    """

    if x not in values:
        x = values[-1]

    binary_encoding = [int(v == x) for v in values]

    return binary_encoding


def get_atom_features(
        atom: Chem.Atom,
        use_chirality: bool = True,
        implicit_hydrogens: bool = True) -> np.ndarray:
    """
    Featurize atom
    
    Args:
        atom (Chem.Atom): Atom 
        use_chirality (bool) 
        implicit_hydrogens (bool)

    Returns:
        atom_feature_vector (np.ndarray)
    """

    allowed_elements = [
        'C', 'N', 'O', 'S', 'F', 'Si', 'P', 'Cl', 'Br', 'Mg', 'Na', 'Ca', 'Fe', 'As', 'Al', 'I',
        'B', 'V', 'K', 'Tl', 'Yb', 'Sb', 'Sn', 'Ag', 'Pd', 'Co', 'Se', 'Ti', 'Zn', 'Li', 'Ge',
        'Cu', 'Au', 'Ni', 'Cd', 'In', 'Mn', 'Zr', 'Cr', 'Pt', 'Hg', 'Pb', 'Unknown'
    ]

    if not implicit_hydrogens:
        allowed_elements = ['H'] + allowed_elements

    # compute atom features

    atom_type_enc = one_hot_encoding(str(atom.GetSymbol()), allowed_elements)

    n_heavy_neighbors_enc = one_hot_encoding(int(atom.GetDegree()), [0, 1, 2, 3, 4, "MoreThanFour"])

    formal_charge_enc = one_hot_encoding(int(atom.GetFormalCharge()), [-3, -2, -1, 0, 1, 2, 3, "Extreme"])

    hybridisation_type_enc = one_hot_encoding(str(atom.GetHybridization()),
                                              ["S", "SP", "SP2", "SP3", "SP3D", "SP3D2", "OTHER"])

    is_in_a_ring_enc = [int(atom.IsInRing())]

    is_aromatic_enc = [int(atom.GetIsAromatic())]

    atomic_mass_scaled = [float((atom.GetMass() - 10.812) / 116.092)]

    vdw_radius_scaled = [float((Chem.GetPeriodicTable().GetRvdw(atom.GetAtomicNum()) - 1.5) / 0.6)]

    covalent_radius_scaled = [float((Chem.GetPeriodicTable().GetRcovalent(atom.GetAtomicNum()) - 0.64) / 0.76)]

    atom_feature_vector = \
        atom_type_enc + n_heavy_neighbors_enc + \
        formal_charge_enc + hybridisation_type_enc + \
        is_in_a_ring_enc + is_aromatic_enc + \
        atomic_mass_scaled + vdw_radius_scaled + \
        covalent_radius_scaled

    if use_chirality:
        chirality_type_enc = one_hot_encoding(str(atom.GetChiralTag()),
                                              ["CHI_UNSPECIFIED", "CHI_TETRAHEDRAL_CW", "CHI_TETRAHEDRAL_CCW",
                                               "CHI_OTHER"])
        atom_feature_vector += chirality_type_enc

    if implicit_hydrogens:
        n_hydrogens_enc = one_hot_encoding(int(atom.GetTotalNumHs()), [0, 1, 2, 3, 4, "MoreThanFour"])
        atom_feature_vector += n_hydrogens_enc

    return np.array(atom_feature_vector)


def get_bond_features(bond: Chem.Bond,
                      use_stereochemistry: bool = True) -> np.ndarray:
    """
    Featurize bond

    Args:
        bond (Chem.Bond): Bond
        use_stereochemistry (bool)

    Returns:
        bond_feature_vector (np.ndarray)
    """

    allowed_bonds = [Chem.rdchem.BondType.SINGLE, Chem.rdchem.BondType.DOUBLE, Chem.rdchem.BondType.TRIPLE,
                     Chem.rdchem.BondType.AROMATIC]

    bond_type_enc = one_hot_encoding(bond.GetBondType(), allowed_bonds)

    bond_is_conj_enc = [int(bond.GetIsConjugated())]

    bond_is_in_ring_enc = [int(bond.IsInRing())]

    bond_feature_vector = bond_type_enc + bond_is_conj_enc + bond_is_in_ring_enc

    if use_stereochemistry:
        stereo_type_enc = one_hot_encoding(str(bond.GetStereo()), ["STEREOZ", "STEREOE", "STEREOANY", "STEREONONE"])
        bond_feature_vector += stereo_type_enc

    return np.array(bond_feature_vector)


def create_pyg_data_lst(x_smiles: Sequence[str], y: Sequence[Number], device: str = 'cpu') -> List[Data]:
    """
    Package a sequence of smiles strings and labels as a list
    of PyTorch geometric data objects, containing the molecule as graph

    Args:
        x_smiles (Sequence[str])
        y (Sequence[Number])
        device (str)

    Returns:
        data_list (List[Data]): List of PyTorch geometric Data objects
    """

    # We use this hack to determine the number of edge and node features
    unrelated_smiles = "O=O"
    unrelated_mol = Chem.MolFromSmiles(unrelated_smiles)
    n_node_features = len(get_atom_features(unrelated_mol.GetAtomWithIdx(0)))
    n_edge_features = len(get_bond_features(unrelated_mol.GetBondBetweenAtoms(0, 1)))

    data_list = []

    for smiles, label in zip(x_smiles, y):

        # convert SMILES to RDKit mol object
        mol = Chem.MolFromSmiles(smiles)

        # get feature dimensions
        n_nodes = mol.GetNumAtoms()
        n_edges = 2 * mol.GetNumBonds()

        # construct node feature matrix X of shape (n_nodes, n_node_features)
        X = np.zeros((n_nodes, n_node_features))

        for atom in mol.GetAtoms():
            X[atom.GetIdx(), :] = get_atom_features(atom)

        X = torch.tensor(X, dtype=torch.float)

        # construct edge index array E of shape (2, n_edges)
        (rows, cols) = np.nonzero(GetAdjacencyMatrix(mol))
        torch_rows = torch.from_numpy(rows.astype(np.int64)).to(torch.long)
        torch_cols = torch.from_numpy(cols.astype(np.int64)).to(torch.long)
        E = torch.stack([torch_rows, torch_cols], dim=0)

        # construct edge feature array EF of shape (n_edges, n_edge_features)
        EF = np.zeros((n_edges, n_edge_features))

        for (k, (i, j)) in enumerate(zip(rows, cols)):
            EF[k] = get_bond_features(mol.GetBondBetweenAtoms(int(i), int(j)))

        EF = torch.tensor(EF, dtype=torch.float)

        # construct label tensor
        y_tensor = torch.tensor(np.array([label]), dtype=torch.float)

        # construct Pytorch Geometric data object and append to data list
        data_list.append(Data(x=X, edge_index=E, edge_attr=EF, y=y_tensor).to(device))

    return data_list



import torch_geometric.nn as geom_nn
from torch import nn

gnn_layer_by_name = {
    "GCN": geom_nn.GCNConv,
    "GAT": geom_nn.GATConv,
    "GraphConv": geom_nn.GraphConv
}


class GNNModel(nn.Module):

    def __init__(
            self, c_in: int, c_hidden: int, c_out: int, num_layers: int = 2, layer_name: str = "GCN",
            dp_rate: float = 0.1, **kwargs):
        """
        GNN model using several possible GNN layers

        Args:
            c_in (int): Dimension of input features
            c_hidden (int): Dimension of hidden features
            c_out (int): Dimension of the output features. Usually number of classes in classification
            num_layers (int): Number of "hidden" graph layers
            layer_name (str): String of the graph layer to use (see gnn_layer_by_name dict)
            dp_rate (float): Dropout rate to apply throughout the network
            **kwargs: Additional arguments for the graph layer (e.g. number of heads for GAT)
        """

        super().__init__()
        gnn_layer = gnn_layer_by_name[layer_name]

        layers = []
        in_channels, out_channels = c_in, c_hidden
        # Build model
        for l_idx in range(num_layers - 1):
            layers += [
                gnn_layer(in_channels=in_channels,
                          out_channels=out_channels,
                          **kwargs),
                nn.ReLU(inplace=True),
                nn.Dropout(dp_rate)
            ]
            in_channels = c_hidden
        layers += [gnn_layer(in_channels=in_channels,
                             out_channels=c_out,
                             **kwargs)]
        self.layers = nn.ModuleList(layers)

    def forward(self, x, edge_index):
        """

        Args:
            x : Input features per node
            edge_index: List of vertex index pairs representing the edges in the graph (PyTorch geometric notation)

        """

        for l in self.layers:
            # For graph layers, we need to add the "edge_index" tensor as additional input
            # All PyTorch Geometric graph layer inherit the class "MessagePassing", hence
            # we can simply check the class type.
            if isinstance(l, geom_nn.MessagePassing):
                x = l(x, edge_index)
            else:
                x = l(x)
        return x



from setuptools import setup, find_packages

setup(
    name='MolGNN',
    version='0.0.2',
    packages=find_packages(),
    url='',
    license='MIT License',
    author='Giulio Mattedi',
    author_email='lxu110@stu.suda.edu.cn',
    description='',
    python_requires='>=3.5, <=3.8',
    
    #install_requires=['rdkit', 'torch', 'numpy', 'torch-geometric', 'torch-scatter', 'torch-sparse']
)
